package com.abi.task.common.tablestore;

import cn.hutool.core.collection.CollectionUtil;
import cn.hutool.core.date.DateUtil;
import com.abi.task.common.api.exception.BusinessException;
import com.abi.task.common.api.exception.ErrorCodeEnum;
import com.abi.task.common.tablestore.common.TableStore;
import com.abi.task.common.tablestore.common.TableStoreEntity;
import com.abi.task.common.utils.IStringUtil;
import com.abi.task.common.utils.TableStoreUtils;
import com.alicloud.openservices.tablestore.model.Column;
import com.alicloud.openservices.tablestore.model.ColumnValue;
import com.alicloud.openservices.tablestore.model.Row;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * 进一步封装的tablestore工具类
 */
@Component
@Slf4j
public class TableStorePlusUtils {

    @Autowired
    private TableStoreUtils tableStoreUtils;

    /**
     * 存储单条 重构方法
     *
     * @param entity
     */
    public void putRow(TableStoreEntity entity) {
        putRow(Arrays.asList(entity));
    }

    /**
     * 存储多行数据（如果主键key相同，则覆盖）
     *
     * @param entityList
     */
    public void putRow(List<TableStoreEntity> entityList) {
        if (CollectionUtil.isEmpty(entityList)) {
            throw new BusinessException(ErrorCodeEnum.ERROR_PARAM.getCode(), "入参列表为空");
        }
        try {
            String primaryKeyName = null;
            String tableName = null;
            List<String> pkValueList = new ArrayList<>();
            List<List<Column>> columnsList = new ArrayList<>();

            for (TableStoreEntity entity : entityList) {
                List<Column> columns = new ArrayList<>();
                //表名
                TableStore tableStore = entity.getClass().getAnnotation(TableStore.class);
                tableName = tableStore.tableName();
                if (StringUtils.isBlank(tableName)) {
                    throw new BusinessException(ErrorCodeEnum.ERROR_PARAM.getCode(), "存储表名为空");
                }
                //主键
                primaryKeyName = tableStore.primaryKeyName();
                if (StringUtils.isBlank(primaryKeyName)) {
                    throw new BusinessException(ErrorCodeEnum.ERROR_PARAM.getCode(), "存储主键为空");
                }
                //主键的驼峰转下划线
                primaryKeyName = IStringUtil.camelToUnderline(primaryKeyName);
                //主键的值
                String primaryKeyVal = null;

                //column列表  循环放入字段
                Field[] fields = entity.getClass().getDeclaredFields();
                for (Field field : fields) {
                    String fieldName = field.getName();
                    String underlineFieldName = IStringUtil.camelToUnderline(fieldName);
                    //获取method
                    Method method = entity.getClass().getMethod("get" + fieldName.substring(0, 1).toUpperCase() + fieldName.substring(1));
                    if (method == null) {
                        throw new BusinessException("get方法不存在");
                    }
                    //反射调用拿到结果
                    Object fieldValue = method.invoke(entity);
                    //如果是主键则赋值,主键值不放入column，不然阿里会报错
                    if (primaryKeyVal == null && underlineFieldName.equals(primaryKeyName)) {
                        primaryKeyVal = (String) fieldValue;
                        continue;
                    }
                    if (fieldValue == null) {
                        continue;
                    }
                    //放入table的column
                    Column column = null;
                    if (fieldValue instanceof String) {
                        column = new Column(underlineFieldName, ColumnValue.fromString(String.valueOf(fieldValue)));
                    } else if (fieldValue instanceof Integer) {
                        column = new Column(underlineFieldName, ColumnValue.fromLong(Long.valueOf((Integer) fieldValue)));
                    } else if (fieldValue instanceof Long) {
                        column = new Column(underlineFieldName, ColumnValue.fromLong((Long) fieldValue));
                    } else if (fieldValue instanceof Double) {
                        column = new Column(underlineFieldName, ColumnValue.fromDouble((Double) fieldValue));
                    } else if (fieldValue instanceof Boolean) {
                        column = new Column(underlineFieldName, ColumnValue.fromBoolean((Boolean) fieldValue));
                    } else if (fieldValue instanceof Date) {
                        column = new Column(underlineFieldName, ColumnValue.fromString(DateUtil.format((Date) fieldValue, "yyyy-MM-dd HH:mm:ss")));
                    } else if (fieldValue instanceof LocalDateTime) {
                        column = new Column(underlineFieldName, ColumnValue.fromString(DateUtil.format((LocalDateTime) fieldValue, "yyyy-MM-dd HH:mm:ss")));
                    } else {
                        log.info("不识别的字段类型 fieldName=" + fieldName);
                        continue;
                    }
                    columns.add(column);
                }
                //放入调用入参
                pkValueList.add(primaryKeyVal);
                columnsList.add(columns);
            }

            //真实去调用
            if (pkValueList.size() > 1) {
                tableStoreUtils.batchPutRow(primaryKeyName, pkValueList, tableName, columnsList);
            } else {
                tableStoreUtils.putRow(primaryKeyName, pkValueList.get(0), tableName, columnsList.get(0));
            }

        } catch (Exception e) {
            log.info("setRow error", e);
            throw new BusinessException("存储table异常");
        }
    }

    /**
     * 读取一行数据
     *
     * @param clz
     * @param primaryKeyVal
     * @param <T>
     * @return
     */
    public <T extends TableStoreEntity> T getRow(Class<T> clz, String primaryKeyVal) {
        T entity = null;

        try {
            //表名
            TableStore tableStore = clz.getAnnotation(TableStore.class);
            String tableName = tableStore.tableName();
            if (StringUtils.isBlank(tableName)) {
                throw new BusinessException("存储表名为空");
            }
            //主键
            String primaryKeyName = tableStore.primaryKeyName();
            if (StringUtils.isBlank(primaryKeyName)) {
                throw new BusinessException("存储主键为空");
            }
            //主键的驼峰转下划线
            primaryKeyName = IStringUtil.camelToUnderline(primaryKeyName);

            //调用tablestore查询到数据
            Row row = tableStoreUtils.getRow(primaryKeyName, primaryKeyVal, tableName);
            //获取字段的值
            Column[] columns = row.getColumns();
            //转化为map
            Map<String, Object> rowMap = new HashMap<>();
            for (Column column : columns) {
                rowMap.put(column.getName(), column.getValue().getValue());
            }

            //转为对象
            entity = clz.newInstance();
            //循环塞值
            Field[] fields = entity.getClass().getDeclaredFields();
            for (Field field : fields) {
                String fieldName = field.getName();
                String underlineFieldName = IStringUtil.camelToUnderline(fieldName);
                Class<?> fieldType = field.getType();
                //拿到tablestore的值
                Object columnVal = rowMap.get(underlineFieldName);
                //转化完成的值
                Object fieldValue = null;
                //获取对象的set方法
                Method method = clz.getMethod("set" + fieldName.substring(0, 1).toUpperCase() + fieldName.substring(1), fieldType);
                //赛主键
                if (underlineFieldName.equals(primaryKeyName)) {
                    fieldValue = primaryKeyVal;
                } else {
                    //塞其他值
                    if (columnVal != null) {
                        if (fieldType == String.class) {
                            fieldValue = (String) columnVal;
                        } else if (fieldType == Integer.class) {
                            fieldValue = Integer.valueOf(String.valueOf(columnVal));
                        } else if (fieldType == Long.class) {
                            fieldValue = (Long) columnVal;
                        } else if (fieldType == Double.class) {
                            fieldValue = (Double) columnVal;
                        } else if (fieldType == Boolean.class) {
                            fieldValue = (Boolean) columnVal;
                        } else if (fieldType == Date.class) {
                            fieldValue = DateUtil.parse((String) columnVal, "yyyy-MM-dd HH:mm:ss");
                        } else if (fieldType == LocalDateTime.class) {
                            fieldValue = DateUtil.parseLocalDateTime((String) columnVal, "yyyy-MM-dd HH:mm:ss");
                        } else {
                            log.info("不识别的字段类型 fieldName=" + fieldName);
                            continue;
                        }
                    }
                }

                //反射调用set
                method.invoke(entity, fieldValue);
            }

        } catch (Exception e) {
            log.info("获取table store异常", e);
            throw new BusinessException("获取table store异常");
        }

        return entity;
    }

    /**
     * 删除行数据
     *
     * @param clz
     * @param primaryKeyVal
     * @param <T>
     */
    public <T extends TableStoreEntity> void deleteRow(Class<T> clz, String primaryKeyVal) {
        try {
            //表名
            TableStore tableStore = clz.getAnnotation(TableStore.class);
            String tableName = tableStore.tableName();
            if (StringUtils.isBlank(tableName)) {
                throw new BusinessException(ErrorCodeEnum.ERROR_PARAM.getCode(), "存储表名为空");
            }
            //主键
            String primaryKeyName = tableStore.primaryKeyName();
            if (StringUtils.isBlank(primaryKeyName)) {
                throw new BusinessException(ErrorCodeEnum.ERROR_PARAM.getCode(), "存储主键为空");
            }
            //主键的驼峰转下划线
            primaryKeyName = IStringUtil.camelToUnderline(primaryKeyName);

            //调用删除
            tableStoreUtils.deleteRow(primaryKeyName, primaryKeyVal, tableName);

        } catch (Exception e) {
            log.info("删除table store异常", e);
            throw new BusinessException("删除table store异常");
        }

    }

}
