package com.abi.qms.platform.advice;



import com.abi.task.common.api.base.BaseResponse;
import com.abi.task.common.api.exception.BusinessException;
import com.abi.task.common.api.exception.ErrorCodeEnum;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.dao.DataAccessException;
import org.springframework.http.converter.HttpMessageNotReadableException;
import org.springframework.util.CollectionUtils;
import org.springframework.validation.BindingResult;
import org.springframework.validation.FieldError;
import org.springframework.web.HttpRequestMethodNotSupportedException;
import org.springframework.web.bind.MethodArgumentNotValidException;
import org.springframework.web.bind.annotation.ExceptionHandler;
import org.springframework.web.bind.annotation.RestControllerAdvice;

import javax.servlet.http.HttpServletRequest;
import java.util.Arrays;
import java.util.List;

/**
 * @author:史良兵
 * @Date:2020/11/19 11:10
 * @Description: 统一异常处理
 */
@RestControllerAdvice
public class ExceptionAdvice {
	
	private Logger logger = LoggerFactory.getLogger(this.getClass());
	
	/**
     * 处理请求对象属性不满足校验规则的异常信息
     *
     * @param request
     * @param exception
     * @return
     * @throws Exception
     */
    @ExceptionHandler(value = {MethodArgumentNotValidException.class})
    public BaseResponse exception(HttpServletRequest request, MethodArgumentNotValidException exception) {
        BindingResult result = exception.getBindingResult();
        List<FieldError> fieldErrors = result.getFieldErrors();
        StringBuilder builder = new StringBuilder();

        if(!CollectionUtils.isEmpty(fieldErrors)){
            builder.append(fieldErrors.get(0).getField()+fieldErrors.get(0).getDefaultMessage()+"!");
        }
        logger.info(Arrays.toString(fieldErrors.toArray()));

        return BaseResponse.createFail(null,ErrorCodeEnum.FAIL.getCode(), builder.toString());
    }

    /**
     * http请求方法不支持
     * @param request
     * @param exception
     * @return
     */
    @ExceptionHandler(value = {HttpRequestMethodNotSupportedException.class})
    public BaseResponse exception(HttpServletRequest request, HttpRequestMethodNotSupportedException exception) {
        String result = exception.getMessage();
        logger.warn(result);
        return BaseResponse.createFail(null, ErrorCodeEnum.ERROR_METHOD_NOT_SUPPORT.getCode(), exception.getMessage());
    }


    /**
     * json错误
     * @param request
     * @param exception
     * @return
     */
    @ExceptionHandler(value = {HttpMessageNotReadableException.class})
    public BaseResponse exception(HttpServletRequest request, HttpMessageNotReadableException exception) {
        String result = exception.getMessage();
        logger.warn(result);
        return BaseResponse.createFail(null, ErrorCodeEnum.ERROR_PARAM.getCode(), exception.getMessage());
    }

    /**
     * 统一处理业务异常
     *
     * @param request
     * @param exception
     * @return
     * @throws Exception
     */
    @ExceptionHandler(value = BusinessException.class)
    public BaseResponse exception(HttpServletRequest request, BusinessException exception) {
        logger.error("业务异常", exception);

        return BaseResponse.createFail(null,exception.getCode(),exception.getMessage());
    }


    /**
     * 数据库异常
     * @param request
     * @param exception
     * @return
     */
    @ExceptionHandler(value = DataAccessException.class)
    public BaseResponse exception(HttpServletRequest request, DataAccessException exception) {
        logger.error("数据库异常", exception);
        return BaseResponse.createFail(null, ErrorCodeEnum.EXCEPT_DB_OPERATION);
    }

    /**
     * 统一处理异常
     *
     * @param request
     * @param exception
     * @return
     * @throws Exception
     */
    @ExceptionHandler(value = Exception.class)
    public BaseResponse exception(HttpServletRequest request, Exception exception) {
        logger.error("系统异常", exception);
        return BaseResponse.createFail(null, ErrorCodeEnum.SERVICE_EXCEPTION);
    }

}
