package com.abi.qms.platform.infrastructure.mq;

import cn.hutool.core.io.file.FileWriter;
import cn.hutool.core.util.ZipUtil;
import com.abi.qms.platform.dao.entity.QrData;
import com.abi.qms.platform.dao.entity.QrInnerData;
import com.abi.qms.platform.dao.entity.QrPackage;
import com.abi.qms.platform.dao.entity.QrRepertorySerialNumber;
import com.abi.qms.platform.dao.enums.QrFormatUrlEnum;
import com.abi.qms.platform.dao.enums.QrPackageGenerateStatusEnum;
import com.abi.qms.platform.dao.enums.QrRepertoryTypeEnum;
import com.abi.qms.platform.dao.mapper.QrPackageMapper;
import com.abi.qms.platform.dao.mapper.QrRepertoryMapper;
import com.abi.qms.platform.dao.mapper.QrRepertorySerialNumberMapper;
import com.abi.qms.platform.dao.vo.result.QrPackageVO;
import com.abi.qms.platform.dao.vo.result.QrRepertoryColumnVO;
import com.abi.qms.platform.dao.vo.result.QrRepertoryVO;
import com.abi.qms.platform.infrastructure.qr.build.parent.SerialBuildCode;
import com.abi.qms.platform.infrastructure.util.BuildCodeUtil;
import com.abi.qms.platform.service.QrDataService;
import com.abi.qms.platform.service.QrPackageService;
import com.baomidou.mybatisplus.core.conditions.query.QueryWrapper;
import com.rabbitmq.client.Channel;
import lombok.SneakyThrows;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.redisson.api.RLock;
import org.redisson.api.RedissonClient;
import org.springframework.amqp.rabbit.annotation.RabbitHandler;
import org.springframework.amqp.rabbit.annotation.RabbitListener;
import org.springframework.amqp.support.AmqpHeaders;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.messaging.Message;
import org.springframework.stereotype.Component;

import java.io.File;
import java.time.LocalDateTime;
import java.util.*;
import java.util.stream.Collectors;

/**
 * 生成码-消费者
 *
 * @author WeiganCai
 * @date: 2021-05-06
 */
@Slf4j
@Component
@RabbitListener(queues = GenerateCodeConsumer.GENERATE_CODE_BY_REPERTORY_QUEUE)
public class GenerateCodeConsumer {

	public static final String GENERATE_CODE_BY_REPERTORY_QUEUE = "generate_code_queue";

	@Autowired
	private QrPackageMapper qrPackageMapper;

	@Autowired
	private QrRepertoryMapper qrRepertoryMapper;

	@Autowired
	private QrRepertorySerialNumberMapper qrRepertorySerialNumberMapper;

	@Autowired
	private QrDataService qrDataService;

	@Autowired
	private QrPackageService qrPackageService;

	@Autowired
	private RedissonClient redissonClient;

	/** 允许单个码生成重复的次数 */
	private static final int REPEAT_BUILD_CODE_TIMES = 10;

	/** 一次生成数量 */
	public static final int ONCE_GENERATE_NUMBER = 5000;

	/** 系统url */
	@Value("${system.url}")
	private String systemUrl;

	@SneakyThrows
	@RabbitHandler
	public void onMessage(QrPackage qrPackage, Message message, Channel channel) {
		try {
			// 生成码
			loopGenerateCode(qrPackage);
		} catch (Exception e) {
			log.error("生成码失败", e);
			// 回滚
			qrPackageService.failedRollback(qrPackage);
		} finally {
			// 手工ack
			Long deliveryTag = (Long)message.getHeaders().get(AmqpHeaders.DELIVERY_TAG);
			channel.basicAck(deliveryTag, false);
		}
	}

	/**
	 * 生成码
	 */
	private void loopGenerateCode(QrPackage qrPackage) {
		// 查询码包关联的幅面关联的码库的信息
		QrPackageVO qrPackageVo = qrPackageMapper.selectQrPackageDetailById(qrPackage.getId());

		// 码数据已生成，return
		if (QrPackageGenerateStatusEnum.GENERATE_SUCCESS.is(qrPackageVo.getGenerateStatus())) {
			return;
		}

		// 更改生成状态为生成中
		QrPackage qrPackageUpdate = new QrPackage().setId(qrPackageVo.getId()).setGenerateStatus(QrPackageGenerateStatusEnum.GENERATING.getCode());
		qrPackageMapper.updateById(qrPackageUpdate);

		// 码库 & 码库的列
		QrRepertoryVO qrRepertory = qrRepertoryMapper.selectQrRepertoryDetailById(qrPackageVo.getQrRepertoryId());
		List<QrRepertoryColumnVO> qrRepertoryColumnList = qrRepertory.getQrRepertoryColumnList();
		Long qrRepertoryId = qrRepertory.getId();
		String deptUrl = qrRepertory.getDeptUrl();

		// 存放每列的url
		Map<Long, String> urlMap = new LinkedHashMap<>();

		// key:qr_repertory_column_id + child_index value:QrRepertorySerialNumber，后续会更新流水号
		// child_index用于标记同一列的第几个码格式变量
		Map<String, QrRepertorySerialNumber> qrRepertorySerialNumberMap = new LinkedHashMap<>();

		// 处理url和流水号
		getUrlAndSerialNumber(qrRepertoryColumnList, qrRepertoryId, deptUrl, urlMap, qrRepertorySerialNumberMap);

		// 申请数量、批次号、创建时间、重复次数
		Long qrNumber = qrPackageVo.getQrNumber();
		String batchNumber = qrPackageVo.getBatchNumber();
		LocalDateTime localDateTime = LocalDateTime.now();
		int repeatTimes = 0;

		// 循环生成码
		loopGenerateCode(qrRepertoryColumnList, qrRepertoryId, urlMap, qrRepertorySerialNumberMap, qrNumber, batchNumber, localDateTime, repeatTimes);

		// 更新码库流水号
		if (!qrRepertorySerialNumberMap.isEmpty()) {
			for (Map.Entry<String, QrRepertorySerialNumber> entry : qrRepertorySerialNumberMap.entrySet()) {
				QrRepertorySerialNumber qrRepertorySerialNumber = entry.getValue();
				QrRepertorySerialNumber updateSerialNumber = new QrRepertorySerialNumber()
						.setId(qrRepertorySerialNumber.getId())
						.setSerialNumber(qrRepertorySerialNumber.getSerialNumber());
				qrRepertorySerialNumberMapper.updateById(updateSerialNumber);
			}
		}

		// 创建码文件
		createCodeFile(qrPackage, qrRepertory);

		// 更改生成状态为已生成
		qrPackageUpdate = new QrPackage().setId(qrPackageVo.getId()).setGenerateStatus(QrPackageGenerateStatusEnum.GENERATE_SUCCESS.getCode());
		qrPackageMapper.updateById(qrPackageUpdate);
	}

	/**
	 * 循环生成码
	 */
	private void loopGenerateCode(List<QrRepertoryColumnVO> qrRepertoryColumnList, Long qrRepertoryId, Map<Long, String> urlMap, Map<String, QrRepertorySerialNumber> qrRepertorySerialNumberMap, Long qrNumber, String batchNumber, LocalDateTime localDateTime, int repeatTimes) {
		// 生成码
		while (true) {
			// 查询已生成的数量
			long alreadyGenerateNumber = qrDataService.selectCountByBatchNumber(batchNumber);

			// 计算还需生成的数量
			long needGenerateNumber = qrNumber - alreadyGenerateNumber;

			if (needGenerateNumber <= 0) {
				log.info("批次号：{}，码生成完毕", batchNumber);
				break;
			}

			// 计算本次最多生成多少码
			long currentMaxGenerateNumber = needGenerateNumber <= ONCE_GENERATE_NUMBER ? needGenerateNumber : ONCE_GENERATE_NUMBER;

			List<QrData> qrDataList = new LinkedList<>();
			// 用Set存放不重复的码
			HashSet<String> uniqueSet = new HashSet<>();
			// 生成码要用的参数
			Map<String, Object> paramMap = new HashMap<>();
			// 码
			StringBuilder codeSb = new StringBuilder();

			for (int i = 0; i < currentMaxGenerateNumber; i++) {
				QrData qrData = new QrData()
						.setQrRepertoryId(qrRepertoryId)
						.setBatchNumber(batchNumber)
						.setQrRepertoryType(QrRepertoryTypeEnum.SYSTEM_GENERATE.getCode())
						.setCreateTime(localDateTime);

				List<QrInnerData> qrInnerDataList = new LinkedList<>();
				for (int j = 0; j < qrRepertoryColumnList.size(); j++) {
					QrRepertoryColumnVO qrRepertoryColumn = qrRepertoryColumnList.get(j);
					Long qrRepertoryColumnId = qrRepertoryColumn.getId();

					// 判断当前列是否为clone列
					Integer cloneColumn = qrRepertoryColumn.getCloneColumn();
					QrInnerData qrInnerData = new QrInnerData();
					if (cloneColumn.compareTo(-1) == 0) {
						// 非clone列，设置code和url
						String code = buildUniqueCode(uniqueSet, qrRepertoryColumn, codeSb, paramMap, qrRepertorySerialNumberMap, 0);
						String url = urlMap.get(qrRepertoryColumnId);
						if (code == null) {
							throw new RuntimeException("生成码失败，重复生成码次数超过" + REPEAT_BUILD_CODE_TIMES + "次");
						}

						qrInnerData.setCode(code).setUrl(url);
						qrInnerData.setQrRepertoryColumnId(qrRepertoryColumnId).setSortNumber(qrRepertoryColumn.getSortNumber());
						qrInnerDataList.add(qrInnerData);
					} else {
						// clone列，设置相同的code和url
						QrInnerData originalQrInnerData = qrInnerDataList.get(cloneColumn);
						qrInnerData.setCode(originalQrInnerData.getCode()).setUrl(originalQrInnerData.getUrl());
						qrInnerData.setQrRepertoryColumnId(qrRepertoryColumnId).setSortNumber(qrRepertoryColumn.getSortNumber());
						qrInnerDataList.add(qrInnerData);
					}
				}

				qrData.setInnerDataList(qrInnerDataList);
				qrDataList.add(qrData);

			}

			try {
				qrDataService.batchInsert(qrDataList);
			} catch (Exception e) {
				// mongoDB唯一索引冲突导致插入失败，继续生成码
				log.error("mongo insert error", e);
				if (++repeatTimes >= REPEAT_BUILD_CODE_TIMES) {
					throw new RuntimeException("生成码失败，重复生成码次数超过" + REPEAT_BUILD_CODE_TIMES + "次");
				}
			}
		}
	}

	/**
	 * 处理url和流水号
	 */
	private void getUrlAndSerialNumber(List<QrRepertoryColumnVO> qrRepertoryColumnList, Long qrRepertoryId, String deptUrl, Map<Long, String> urlMap, Map<String, QrRepertorySerialNumber> qrRepertorySerialNumberMap) {
		for (int i = 0; i < qrRepertoryColumnList.size(); i++) {
			QrRepertoryColumnVO qrRepertoryColumn = qrRepertoryColumnList.get(i);
			Long qrRepertoryColumnId = qrRepertoryColumn.getId();

			// 每列的url格式
			Integer urlType = qrRepertoryColumn.getUrlType();
			String url = "";
			if (QrFormatUrlEnum.SYSTEM_URL.is(urlType)) {
				url = systemUrl;
			} else if (QrFormatUrlEnum.DEPT_URL.is(urlType)) {
				url = deptUrl;
			}
			// 如果url没有以/结尾，则加上/
			if (StringUtils.isNotBlank(url) && !url.endsWith("/")) {
				url += "/";
			}
			urlMap.put(qrRepertoryColumnId, url);

			// 使用到流水号参数的列
			String[] buildClassArr = qrRepertoryColumn.getQrFormatVO().getCodeVariableBuildClass().split(",");
			for (int j = 0; j < buildClassArr.length; j++) {
				String buildClass = buildClassArr[j];
				if (buildClass.startsWith(SerialBuildCode.SERIAL_START_WITH)) {
					// 查询码库流水号，若存在，则流水号从记录值开始；若不存在，从0开始
					QueryWrapper<QrRepertorySerialNumber> qrRepertorySerialNumberQw = new QueryWrapper<>();
					qrRepertorySerialNumberQw.eq("qr_repertory_column_id", qrRepertoryColumnId);
					qrRepertorySerialNumberQw.eq("child_index", j);

					// 加锁查询插入流水号
					RLock lock = redissonClient.getLock("serial_number_lock_" + qrRepertoryColumnId + "_" + j);
					QrRepertorySerialNumber qrRepertorySerialNumber = null;
					try {
						lock.lock();
						qrRepertorySerialNumber = qrRepertorySerialNumberMapper.selectOne(qrRepertorySerialNumberQw);
						if (qrRepertorySerialNumber == null) {
							qrRepertorySerialNumber = new QrRepertorySerialNumber()
									.setQrRepertoryId(qrRepertoryId)
									.setQrRepertoryColumnId(qrRepertoryColumnId)
									.setChildIndex(j)
									.setSerialBuildClass(buildClass)
									.setSerialNumber(0L);
							qrRepertorySerialNumberMapper.insert(qrRepertorySerialNumber);
						}
					} finally {
						lock.unlock();
					}

					// key:qr_repertory_column_id + child_index
					qrRepertorySerialNumberMap.put(qrRepertoryColumnId + "_" + j, qrRepertorySerialNumber);
				}
			}

		}
	}

	/**
	 * 创建码文件
	 */
	private void createCodeFile(QrPackage qrPackage, QrRepertoryVO qrRepertory) {
		String batchNumber = qrPackage.getBatchNumber();
		List<QrData> qrDataList = qrDataService.selectCodeByBatch(batchNumber);
		List<QrRepertoryColumnVO> qrRepertoryColumnList = qrRepertory.getQrRepertoryColumnList();
		StringBuilder content = new StringBuilder();
		// 第一行为每列的名称，用“,”分隔
		String dataComposition = qrRepertoryColumnList.stream().map(col -> col.getAlias()).collect(Collectors.joining(","));
		content.append(dataComposition).append("\r\n");

		// 拼接码
		List<QrInnerData> innerDataList;
		String innerData;
		for (QrData qrData : qrDataList) {
			innerDataList = qrData.getInnerDataList();
			innerData = innerDataList.stream().map(qrInnerData -> qrInnerData.getUrl() + qrInnerData.getCode()).collect(Collectors.joining(","));
			content.append(innerData).append("\r\n");
		}

		// 生成txt和zip文件
		FileWriter fileWriter = new FileWriter(batchNumber + ".txt");
		fileWriter.write(content.toString());
		File txtFile = fileWriter.getFile();
		File zipFile = ZipUtil.zip(txtFile);

		// 删除生成的临时文件，上传阿里云，并更新码包的下载路径
		qrPackageService.delFileAndUpdatePath(txtFile, zipFile, qrPackage.getId());

	}

	/**
	 * 生成不重复的码
	 */
	private String buildUniqueCode( HashSet<String> uniqueSet, QrRepertoryColumnVO qrRepertoryColumn, StringBuilder codeSb, Map<String, Object> paramMap, Map<String, QrRepertorySerialNumber> qrRepertorySerialNumberMap, int buildTimes) {
		// 生成码
		buildCodeByClass(qrRepertoryColumn, codeSb, paramMap, qrRepertorySerialNumberMap);
		String code = codeSb.toString();

		// add成功，说明码没有重复
		if (uniqueSet.add(code)) {
			return code;
		}

		// 单个码重复次数大于上限
		if (buildTimes >= REPEAT_BUILD_CODE_TIMES) {
			return null;
		}

		// 码重复，递归执行下一次
		return buildUniqueCode(uniqueSet, qrRepertoryColumn, codeSb, paramMap, qrRepertorySerialNumberMap, ++buildTimes);
	}

	/**
	 * 清空codeSb，调用buildSingleCode方法生成码
	 */
	private void buildCodeByClass(QrRepertoryColumnVO qrRepertoryColumn, StringBuilder codeSb, Map<String, Object> paramMap, Map<String, QrRepertorySerialNumber> qrRepertorySerialNumberMap) {
		// 使用一个对象，删除内容，不每次都新建一个对象
		codeSb.delete(0, codeSb.length());

		String[] buildClassArr = qrRepertoryColumn.getQrFormatVO().getCodeVariableBuildClass().split(",");

		for (int j = 0; j < buildClassArr.length; j++) {
			String buildClass = buildClassArr[j];
			String code;
			if (buildClass.startsWith(SerialBuildCode.SERIAL_START_WITH)) {
				QrRepertorySerialNumber qrRepertorySerialNumber = qrRepertorySerialNumberMap.get(qrRepertoryColumn.getId() + "_" + j);
				// 查询当前流水号的值，放入map中
				paramMap.put(buildClass, qrRepertorySerialNumber.getSerialNumber());
				code = BuildCodeUtil.buildSingleCode(buildClass, paramMap);
				// 重新赋值
				qrRepertorySerialNumber.setSerialNumber(Long.parseLong(code));
			} else {
				code = BuildCodeUtil.buildSingleCode(buildClass, paramMap);
			}
			codeSb.append(code);
		}
	}

}
