package com.abi.qms.platform.service.impl;

import cn.hutool.core.io.FileUtil;
import cn.hutool.core.io.IoUtil;
import com.abi.qms.platform.dao.entity.QrPackage;
import com.abi.qms.platform.dao.entity.QrPackageDownloadRecord;
import com.abi.qms.platform.dao.enums.QrPackageApplyStatusEnum;
import com.abi.qms.platform.dao.enums.QrPackageGenerateStatusEnum;
import com.abi.qms.platform.dao.enums.QrRepertoryTypeEnum;
import com.abi.qms.platform.dao.mapper.QrPackageDownloadRecordMapper;
import com.abi.qms.platform.dao.mapper.QrPackageMapper;
import com.abi.qms.platform.dao.mapper.QrRepertoryMapper;
import com.abi.qms.platform.dao.vo.result.QrPackageVO;
import com.abi.qms.platform.dto.req.*;
import com.abi.qms.platform.dto.res.GetQrPackageDetailRes;
import com.abi.qms.platform.dto.res.ListQrPackageRes;
import com.abi.qms.platform.infrastructure.mq.GenerateCodeConsumer;
import com.abi.qms.platform.infrastructure.util.AssertUtil;
import com.abi.qms.platform.infrastructure.util.PageUtil;
import com.abi.qms.platform.infrastructure.util.UserUtil;
import com.abi.qms.platform.service.QrDataService;
import com.abi.qms.platform.service.QrPackageService;
import com.abi.task.common.api.exception.BusinessException;
import com.abi.task.common.utils.PojoConverterUtils;
import com.baomidou.mybatisplus.core.conditions.query.QueryWrapper;
import com.baomidou.mybatisplus.core.metadata.IPage;
import com.baomidou.mybatisplus.core.toolkit.IdWorker;
import lombok.extern.slf4j.Slf4j;
import org.springframework.amqp.core.AmqpTemplate;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpHeaders;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import javax.servlet.http.HttpServletResponse;
import java.io.*;
import java.net.URLEncoder;
import java.util.List;

/**
 * 码包 Service业务层处理
 *
 * @author WeiganCai
 * @date 2021-04-28
 */
@Service
@Slf4j
public class QrPackageServiceImpl implements QrPackageService {

	@Autowired
	private QrPackageMapper qrPackageMapper;

	@Autowired
	private AmqpTemplate amqpTemplate;

	@Autowired
	private QrDataService qrDataService;

	@Autowired
	private QrPackageDownloadRecordMapper qrPackageDownloadRecordMapper;

	/**
	 * 保存码包
	 */
	@Override
	@Transactional(rollbackFor = Exception.class)
	public void saveQrPackage(SaveQrPackageReq req) {
		Long qrPackageId = req.getId();
		// 修改码包时的条件限制
		QrPackage qrPackage;
		if (qrPackageId != null) {
			qrPackage = qrPackageMapper.selectById(qrPackageId);
			AssertUtil.isNull(qrPackage, "码包不存在");
			// 只有待审核、审核拒绝状态下才可编辑
			Integer applyStatus = qrPackage.getApplyStatus();
			if (QrPackageApplyStatusEnum.REVIEW_PASS.is(applyStatus)) {
				throw new BusinessException("码包不可编辑");
			}
		}

		// 码包名称不可重复
		QueryWrapper<QrPackage> qrPackageQw = new QueryWrapper<>();
		qrPackageQw.eq("name", req.getName());
		qrPackageQw.eq("is_delete", 0);
		if (qrPackageId != null) {
			qrPackageQw.ne("id", qrPackageId);
		}
		Integer count = qrPackageMapper.selectCount(qrPackageQw);
		if (!count.equals(0)) {
			throw new BusinessException("码包名称已存在");
		}

		// copy属性值
		qrPackage = PojoConverterUtils.copy(req, QrPackage.class);

		// 新增or修改
		Long userId = UserUtil.getUser().getId();
		qrPackage.setUpdateBy(userId);
		if (qrPackageId != null) {
			// 修改时，需要重置码包的状态为待审核
			qrPackage.setApplyStatus(QrPackageApplyStatusEnum.WAIT_REVIEW.getCode());
			qrPackageMapper.updateById(qrPackage);
		} else {
			qrPackage.setBatchNumber(IdWorker.getIdStr());
			qrPackage.setCreateBy(userId);
			qrPackageMapper.insert(qrPackage);
		}

	}

	/**
	 * 码包分页查询
	 */
	@Override
	public ListQrPackageRes listQrPackage(ListQrPackageReq req) {
		// 分页查询
		IPage<QrPackageVO> iPage = qrPackageMapper.listQrPackage(PageUtil.createPage(req), req);
		List<QrPackageVO> qrPackageList = iPage.getRecords();

		// 封装出参、放入分页信息
		ListQrPackageRes res = new ListQrPackageRes();
		PageUtil.copyPageInfo(res, iPage);
		List<ListQrPackageRes.QrPackageBean> qrPackageBeanList = PojoConverterUtils.copyList(qrPackageList, ListQrPackageRes.QrPackageBean.class);
		res.setQrRepertoryBeanList(qrPackageBeanList);

		return res;
	}

	/**
	 * 码包查询详情
	 */
	@Override
	public GetQrPackageDetailRes getQrPackageDetail(GetQrPackageDetailReq req) {
		// 查询码包对象
		QrPackageVO qrPackage = qrPackageMapper.selectQrPackageDetailById(req.getId());
		AssertUtil.isNull(qrPackage, "码包不存在");

		return PojoConverterUtils.copy(qrPackage, GetQrPackageDetailRes.class);
	}

	/**
	 * 删除码包
	 */
	@Override
	@Transactional(rollbackFor = Exception.class)
	public void deleteQrPackage(DeleteQrPackageReq req) {
		QrPackage qrPackage = qrPackageMapper.selectById(req.getId());
		AssertUtil.isNull(qrPackage, "码包不存在");
		// 只有待审核、审核拒绝状态下才可删除
		Integer applyStatus = qrPackage.getApplyStatus();
		if (QrPackageApplyStatusEnum.REVIEW_PASS.is(applyStatus)) {
			throw new BusinessException("码包不可删除");
		}

		// 逻辑删除
		QrPackage update = new QrPackage().setId(req.getId()).setIsDelete(1);
		qrPackageMapper.updateById(update);
	}

	/**
	 * 审核通过码包
	 */
	@Override
	@Transactional(rollbackFor = Exception.class)
	public void passQrPackage(PassQrPackageReq req) {
		QrPackageVO qrPackage = qrPackageMapper.selectQrPackageDetailById(req.getId());
		AssertUtil.isNull(qrPackage, "码包不存在");

		// 只有待审核、审核拒绝状态下才可通过
		Integer applyStatus = qrPackage.getApplyStatus();
		if (QrPackageApplyStatusEnum.REVIEW_PASS.is(applyStatus)) {
			throw new BusinessException("码包不可通过");
		}

		QrPackage update = new QrPackage().setId(req.getId()).setApplyStatus(QrPackageApplyStatusEnum.REVIEW_PASS.getCode());
		qrPackageMapper.updateById(update);

		// 码库类型为系统生成时，放入rabbitmq，生成码
		if (QrRepertoryTypeEnum.SYSTEM_GENERATE.is(qrPackage.getQrRepertoryType())) {
			amqpTemplate.convertAndSend(GenerateCodeConsumer.GENERATE_CODE_BY_REPERTORY_QUEUE, qrPackage);
		}
	}

	/**
	 * 审核拒绝码包
	 */
	@Override
	@Transactional(rollbackFor = Exception.class)
	public void refuseQrPackage(RefuseQrPackageReq req) {
		QrPackage qrPackage = qrPackageMapper.selectById(req.getId());
		AssertUtil.isNull(qrPackage, "码包不存在");

		// 只有待审核、审核拒绝状态下才可拒绝
		Integer applyStatus = qrPackage.getApplyStatus();
		if (QrPackageApplyStatusEnum.REVIEW_PASS.is(applyStatus)) {
			throw new BusinessException("码包不可拒绝");
		}

		QrPackage update = new QrPackage().setId(req.getId()).setApplyStatus(QrPackageApplyStatusEnum.REVIEW_REFUSE.getCode());
		qrPackageMapper.updateById(update);
	}

	@Override
	public void downloadQrPackage(DownloadQrPackageReq req, HttpServletResponse response) {
		QrPackage qrPackage = qrPackageMapper.selectById(req.getId());
		AssertUtil.isNull(qrPackage, "码包不存在");

		// 只有已生成状态下才可下载
		Integer generateStatus = qrPackage.getGenerateStatus();
		if (!QrPackageGenerateStatusEnum.GENERATE_SUCCESS.is(generateStatus)) {
			throw new BusinessException("码包不可下载");
		}

		BufferedInputStream in = null;
		BufferedOutputStream out = null;
		try {
			// TODO 从阿里云OSS读取文件，替换下面的测试代码
			// TODO in = new BufferedInputStream(ossObject.getObjectContent());
			in = new BufferedInputStream(new FileInputStream(new File("E:\\firefox_download\\202104161027287567357.zip")));
			out = new BufferedOutputStream(response.getOutputStream());

			String fileName = qrPackage.getBatchNumber() + ".zip";
			response.setHeader("Content-Disposition","attachment;filename=" + URLEncoder.encode(fileName,"utf-8"));

			byte [] b = new byte[1024];
			int len = 0;
			while(-1 != (len = in.read(b))) {
				out.write(b, 0, len);
			}
			out.flush();
		} catch (IOException e) {
			log.error("IOException", e);
		} finally {
			IoUtil.close(in);
			IoUtil.close(out);
		}

		// 插入码包下载记录
		Long userId = UserUtil.getUser().getId();
		QrPackageDownloadRecord downloadRecord = new QrPackageDownloadRecord()
				.setCreateBy(userId)
				.setQrPackageId(qrPackage.getId());
		qrPackageDownloadRecordMapper.insert(downloadRecord);

	}

	/**
	 * 删除生成的临时文件，并更新码包的下载路径
	 * @param txtFile txt文件
	 * @param zipFile 压缩文件
	 */
	@Override
	public void delFileAndUpdatePath(File txtFile, File zipFile, Long qrPackageId) {
		// TODO 上传zip包到阿里云OSS后，删除本次生成的txt和zip文件
		FileUtil.del(txtFile);
		FileUtil.del(zipFile);

		// TODO 更新下载路径
		String downloadPath = "";
		QrPackage update = new QrPackage().setId(qrPackageId).setDownloadPath(downloadPath);
		qrPackageMapper.updateById(update);
	}

	/**
	 * 码生成失败回滚
	 */
	@Override
	public void failedRollback(QrPackage qrPackage) {
		// 删除本次生成的码
		qrDataService.deleteCodeByBatchNumber(qrPackage.getBatchNumber());

		// 更改生成状态为生成失败
		QrPackage qrPackageUpdate = new QrPackage().setId(qrPackage.getId()).setGenerateStatus(QrPackageGenerateStatusEnum.GENERATE_FAIL.getCode());
		qrPackageMapper.updateById(qrPackageUpdate);
	}
}
